import glob
import os

import PIL
import matplotlib.pyplot as plt
import numpy as np
import scipy.io as sio
import torch
from torch.utils.data.dataset import Dataset
from torchvision.datasets.utils import download_and_extract_archive


class Cars(Dataset):
    def __init__(self, root: str, download: bool = False, train: bool = True):
        """

        :param root: Root directory of dataset
        :param download: If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        :param train:
        """
        self.resources = [
            "http://www.scottreed.info/files/nips2015-analogy-data.tar.gz"
        ]
        self.root = root
        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )
        if train:
            self.dataset = dict(
                view_1=torch.load(os.path.join(self.raw_folder, "view_1.pt")),
                view_2=torch.load(os.path.join(self.raw_folder, "view_2.pt")),
            )
        else:
            pass

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def __getitem__(self, index):
        return {
            "views": (self.dataset["view_1"][index].astype(np.float32), self.dataset["view_2"][index].astype(np.float32)),
            "index": index,
        }

    def __len__(self):
        return len(self.dataset["view_1"])

    def _check_raw_exists(self) -> bool:
        return os.path.exists(
            os.path.join(self.raw_folder, "nips2015-analogy-data.tar.gz")
        )

    def _check_exists(self) -> bool:
        return os.path.exists(
            os.path.join(self.raw_folder, "view_1.pt")
        ) and os.path.exists(os.path.join(self.raw_folder, "view_2.pt"))

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if not self._check_raw_exists():
            os.makedirs(self.raw_folder, exist_ok=True)
            import ssl

            ssl._create_default_https_context = ssl._create_unverified_context
            # download files
            for url in self.resources:
                filename = url.rpartition("/")[2]
                download_and_extract_archive(
                    url, download_root=self.raw_folder, filename=filename
                )
        if self._check_exists():
            return
        print("Processing...")
        view_1, view_2 = get_cars3d(os.path.join(self.raw_folder, "data/cars/"))
        with open(os.path.join(self.raw_folder, "view_1.pt"), "wb") as f:
            torch.save(view_1, f)
        with open(os.path.join(self.raw_folder, "view_2.pt"), "wb") as f:
            torch.save(view_2, f)
        print("Done!")


# Shuffle the private information
def sample(n1=183 * 2, n2=24):
    idx = []
    for i in range(n1):
        idx.append(np.random.permutation(n2) + i * n2)

    return np.hstack(idx)


# Get the Car3D dataset
def get_cars3d(filedir="./data/cars/"):
    tmp = []
    for f in glob.glob(filedir + "*mesh.mat"):
        a = sio.loadmat(f)
        tt = np.zeros((a["im"].shape[3], a["im"].shape[4], 64, 64, 3))
        for i in range(a["im"].shape[3]):
            for j in range(a["im"].shape[4]):
                pic = PIL.Image.fromarray(a["im"][:, :, :, i, j])
                pic.thumbnail((64, 64), PIL.Image.ANTIALIAS)
                tt[i, j, :, :, :] = np.array(pic) / 255.0

        b = torch.tensor(tt)
        c = b.permute(0, 1, 4, 2, 3)
        tmp.append(c)

    data = torch.stack(tmp, dim=0)

    imgs = data.numpy()
    imgs = np.transpose(imgs, (2, 0, 1, 3, 4, 5))

    # 4 elevations
    elv1 = imgs[0]
    elv2 = imgs[1]
    elv3 = imgs[2]
    elv4 = imgs[3]

    # Show samples
    fig, axarr = plt.subplots(1, 4)
    axarr[0].imshow(np.transpose(elv1[0, 1, :, :, :], (1, 2, 0)))
    axarr[0].axis("off")
    axarr[1].imshow(np.transpose(elv2[0, 7, :, :, :], (1, 2, 0)))
    axarr[1].axis("off")
    axarr[2].imshow(np.transpose(elv3[0, 14, :, :, :], (1, 2, 0)))
    axarr[2].axis("off")
    axarr[3].imshow(np.transpose(elv4[0, 21, :, :, :], (1, 2, 0)))
    axarr[3].axis("off")

    fig.suptitle(
        "Samples of both views, left two for view1, right two for view2", fontsize=10
    )
    plt.show()

    # Get the two views in order
    view1 = []
    view2 = []

    for i in range(elv1.shape[0]):
        # Lower elevations
        view1.append(elv1[i, :, :, :, :])
        view1.append(elv2[i, :, :, :, :])
        # Higher elevations
        view2.append(elv3[i, :, :, :, :])
        view2.append(elv4[i, :, :, :, :])

    view1 = np.concatenate(view1, axis=0)
    view2 = np.concatenate(view2, axis=0)

    return view1, view2



"""
Processing of Twitter Data:

MIT License

Copyright (c) 2016 Adrian Benton

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
"""

import gzip
import os

import numpy as np
from torch.utils.data.dataset import Dataset
from torchvision.datasets.utils import download_and_extract_archive


class Twitter(Dataset):
    def __init__(
        self, root, viewstokeep=None, download=False, maxrows=-1, replaceempty=True
    ):
        self.resources = [
            "https://www.cs.jhu.edu/~mdredze/datasets/multiview_embeddings/user_6views_tfidf_pcaEmbeddings_userTweets+networks.tsv.gz",
            "https://www.cs.jhu.edu/~mdredze/datasets/multiview_embeddings/friend_and_hashtag_prediction_userids.zip",
        ]
        self.root = root
        if download:
            self.download()
        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )

        ids, views = ldViews(
            os.path.join(
                self.raw_folder,
                "user_6views_tfidf_pcaEmbeddings_userTweets+networks.tsv",
            ),
            viewstokeep,
            replaceempty=replaceempty,
            maxrows=maxrows,
        )
        self.dataset = dict(ids=ids, views=views)

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def __getitem__(self, index):
        batch = {"index": index.astype(np.float32)}
        batch["userid"] = self.ids[index].astype(np.float32)
        batch["views"] = [view[index].astype(np.float32) for view in self.views]
        return batch

    def __len__(self):
        return len(self.dataset["views"][0])

    def _check_raw_exists(self) -> bool:
        return os.path.exists(
            os.path.join(
                self.raw_folder,
                "user_6views_tfidf_pcaEmbeddings_userTweets+networks.tsv.gz",
            )
        ) and os.path.exists(
            os.path.join(self.raw_folder, "friend_and_hashtag_prediction_userids.zip")
        )

    def _check_exists(self) -> bool:
        return os.path.exists(
            os.path.join(
                self.raw_folder,
                "user_6views_tfidf_pcaEmbeddings_userTweets+networks.tsv",
            )
        )

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if not self._check_raw_exists():
            os.makedirs(self.raw_folder, exist_ok=True)
            import ssl

            ssl._create_default_https_context = ssl._create_unverified_context
            # download files
            for url in self.resources:
                filename = url.rpartition("/")[2]
                download_and_extract_archive(
                    url, download_root=self.raw_folder, filename=filename
                )


def fopen(p, flag="r"):
    """Opens as gzipped if appropriate, else as ascii."""
    if p.endswith(".gz"):
        if "w" in flag:
            return gzip.open(p, "wb")
        else:
            return gzip.open(p, "rt")
    else:
        return open(p, flag)


def ldViews(inPath, viewstokeep, replaceempty=True, maxrows=-1):
    """
    Loads dataset for each view.
    Input file is tab-separated: first column ID, next k columns are number of documents that
    go into each view, other columns are the views themselves.  Features in each view are
    space-separated floats -- dense.

    replaceempty: If true, then replace empty rows with the mean embedding for each view

    Returns list of IDs, and list with an (N X F) matrix for each view.
    """

    N = 0  # Number of examples

    V = 0  # Number of views
    FperV = []  # Number of features per view

    f = fopen(inPath)
    flds = f.readline().split("\t")

    V = int((len(flds) - 1) / 2)

    # Use all views
    if not viewstokeep:
        viewstokeep = [i for i in range(V)]

    for fld in flds[1 + V :]:
        FperV.append(len(fld.split()))
    f.close()

    f = fopen(inPath)

    flds = f.readline().strip().split("\t")

    F = [len(fld.split()) for fld in flds[(1 + V) :]]
    N += 1

    # Only reduce to kept views
    F = [F[v] for v in viewstokeep]

    for ln in f:
        N += 1
    f.close()

    data = [np.zeros((maxrows, numFs)) for numFs in F]
    ids = []

    f = fopen(inPath)
    for lnidx, ln in enumerate(f):
        if (maxrows > 0) and (lnidx >= maxrows):
            break

        if not lnidx % 10000:
            print("Reading line: %dK" % (lnidx / 1000))

        flds = ln.split("\t")
        ids.append(int(flds[0]))

        viewstrs = flds[(1 + V) :]

        for idx, viewstr in enumerate(viewstrs):
            if idx not in viewstokeep:
                continue

            idx = viewstokeep.index(idx)

            for fidx, v in enumerate(viewstr.split()):
                data[idx][lnidx, fidx] = float(v)

    f.close()

    # Replace empty rows with the mean for each view
    if replaceempty:
        means = [
            np.sum(d, axis=0) / np.sum(1.0 * (np.abs(d).sum(axis=1) > 0.0))
            for d in data
        ]
        for i in range(maxrows):
            for nvIdx, vIdx in enumerate(viewstokeep):
                if np.sum(data[nvIdx][i, :]) == 0.0:
                    data[nvIdx][i, :] = means[nvIdx]

    return ids, data


def ldK(p, viewstokeep):
    """
    Returns matrix K, indicating which (example, view) pair is missing.

    p: Path to data file
    viewstokeep: Indices of views we want to keep.
    """

    numLns = 0

    f = fopen(p)
    for ln in f:
        numviews = ln.count("\t") / 2
        numLns += 1
    f.close()

    # Use all views
    if not viewstokeep:
        viewstokeep = [i for i in range(numviews)]

    K = np.ones((numLns, len(viewstokeep)))

    # We keep count of number of tweets and infos collected in each view, and first field is the user ID.
    # Just zero out those views where we did not collect any data for that view
    f = fopen(p)
    for lnIdx, ln in enumerate(f):
        flds = ln.split("\t")
        for idx, vidx in enumerate(viewstokeep):
            count = int(flds[vidx + 1])
            if count < 1:
                K[lnIdx, idx] = 0.0
    f.close()

    return K



import io
import json
import os
import pickle
from collections import defaultdict, Counter, OrderedDict

import numpy as np
import torch
import torchvision
from nltk.tokenize import sent_tokenize, word_tokenize
from torch.utils.data.dataset import Dataset
from torchvision import datasets
from torchvision.datasets.utils import download_and_extract_archive
from tqdm import tqdm


class Cub(Dataset):
    def __init__(
            self, root, download=False, train=True, image_tx=None,
    ):
        self.resources = [
            "http://www.robots.ox.ac.uk/~yshi/mmdgm/datasets/cub.zip",
        ]
        self.root = root
        if download:
            self.download()
        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )
        if image_tx is None:
            image_tx=torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor()]
            )
        if train:
            self.images = datasets.ImageFolder(os.path.join(self.raw_folder, "cub/train"), transform=image_tx)
        else:
            self.images = datasets.ImageFolder(os.path.join(self.raw_folder, "cub/test"), transform=image_tx)
        tx = lambda data: torch.Tensor(data)
        self.sentences = CUBSentences(self.raw_folder, train=train, transform=tx, max_sequence_length=32)

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def __getitem__(self, index):
        batch = {"index": index}
        image, label = self.images[index//10]
        batch["image"] = image
        batch["label"] = label
        batch["sentence"] = self.sentences[index][0]
        return batch

    def __len__(self):
        return len(self.sentences)

    def _check_raw_exists(self) -> bool:
        return os.path.exists(
            os.path.join(
                self.raw_folder,
                "cub.zip",
            )
        )

    def _check_exists(self) -> bool:
        return os.path.exists(
            os.path.join(
                self.raw_folder,
                "cub",
            )
        )

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if not self._check_raw_exists():
            os.makedirs(self.raw_folder, exist_ok=True)
            import ssl

            ssl._create_default_https_context = ssl._create_unverified_context
            # download files
            for url in self.resources:
                filename = url.rpartition("/")[2]
                download_and_extract_archive(
                    url, download_root=self.raw_folder, filename=filename
                )


class OrderedCounter(Counter, OrderedDict):
    """Counter that remembers the order elements are first encountered."""

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)


class CUBSentences(Dataset):

    def __init__(self, root_data_dir, train=True, transform=None, **kwargs):
        super().__init__()
        self.data_dir = os.path.join(root_data_dir, 'cub')
        self.train = train
        self.max_sequence_length = kwargs.get('max_sequence_length', 32)
        self.min_occ = kwargs.get('min_occ', 3)
        self.transform = transform
        os.makedirs(os.path.join(root_data_dir, "lang_emb"), exist_ok=True)

        self.gen_dir = os.path.join(self.data_dir, "oc{}_msl{}".
                                    format(self.min_occ, self.max_sequence_length))
        if train:
            self.raw_data_path = os.path.join(self.data_dir, 'text_trainvalclasses.txt')
        else:
            self.raw_data_path = os.path.join(self.data_dir, 'text_testclasses.txt')

        os.makedirs(self.gen_dir, exist_ok=True)
        self.data_file = 'cub.{}.s{}'.format("train" if self.train else "test", self.max_sequence_length)
        self.vocab_file = 'cub.vocab'

        if not os.path.exists(os.path.join(self.gen_dir, self.data_file)):
            print("Data file not found for {} split at {}. Creating new... (this may take a while)".format(
                "train" if self.train else "test", os.path.join(self.gen_dir, self.data_file)))
            self._create_data()

        else:
            self._load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        sent = self.data[str(idx)]['idx']
        if self.transform is not None:
            sent = self.transform(sent)
        return sent, self.data[str(idx)]['length']

    @property
    def vocab_size(self):
        return len(self.w2i)

    @property
    def pad_idx(self):
        return self.w2i['<pad>']

    @property
    def eos_idx(self):
        return self.w2i['<eos>']

    @property
    def unk_idx(self):
        return self.w2i['<unk>']

    def get_w2i(self):
        return self.w2i

    def get_i2w(self):
        return self.i2w

    def _load_data(self, vocab=True):
        with open(os.path.join(self.gen_dir, self.data_file), 'rb') as file:
            self.data = json.load(file)

        if vocab:
            self._load_vocab()

    def _load_vocab(self):
        if not os.path.exists(os.path.join(self.gen_dir, self.vocab_file)):
            self._create_vocab()
        with open(os.path.join(self.gen_dir, self.vocab_file), 'r') as vocab_file:
            vocab = json.load(vocab_file)
        self.w2i, self.i2w = vocab['w2i'], vocab['i2w']

    def _create_data(self):
        if self.train and not os.path.exists(os.path.join(self.gen_dir, self.vocab_file)):
            self._create_vocab()
        else:
            self._load_vocab()

        with open(self.raw_data_path, 'r') as file:
            text = file.read()
            sentences = sent_tokenize(text)

        data = defaultdict(dict)
        pad_count = 0

        for i, line in enumerate(tqdm(sentences)):
            words = word_tokenize(line)

            tok = words[:self.max_sequence_length - 1]
            tok = tok + ['<eos>']
            length = len(tok)
            if self.max_sequence_length > length:
                tok.extend(['<pad>'] * (self.max_sequence_length - length))
                pad_count += 1
            idx = [self.w2i.get(w, self.w2i['<exc>']) for w in tok]

            id = len(data)
            data[id]['tok'] = tok
            data[id]['idx'] = idx
            data[id]['length'] = length

        print("{} out of {} sentences are truncated with max sentence length {}.".
              format(len(sentences) - pad_count, len(sentences), self.max_sequence_length))
        with io.open(os.path.join(self.gen_dir, self.data_file), 'wb') as data_file:
            data = json.dumps(data, ensure_ascii=False)
            data_file.write(data.encode('utf8', 'replace'))

        self._load_data(vocab=False)

    def _create_vocab(self):

        assert self.train, "Vocabulary can only be created for training file."

        with open(self.raw_data_path, 'r') as file:
            text = file.read()
            sentences = sent_tokenize(text)

        occ_register = OrderedCounter()
        w2i = dict()
        i2w = dict()

        special_tokens = ['<exc>', '<pad>', '<eos>']
        for st in special_tokens:
            i2w[len(w2i)] = st
            w2i[st] = len(w2i)

        texts = []
        unq_words = []

        for i, line in enumerate(sentences):
            words = word_tokenize(line)
            occ_register.update(words)
            texts.append(words)

        for w, occ in occ_register.items():
            if occ > self.min_occ and w not in special_tokens:
                i2w[len(w2i)] = w
                w2i[w] = len(w2i)
            else:
                unq_words.append(w)

        assert len(w2i) == len(i2w)

        print("Vocabulary of {} keys created, {} words are excluded (occurrence <= {})."
              .format(len(w2i), len(unq_words), self.min_occ))

        vocab = dict(w2i=w2i, i2w=i2w)
        with io.open(os.path.join(self.gen_dir, self.vocab_file), 'wb') as vocab_file:
            data = json.dumps(vocab, ensure_ascii=False)
            vocab_file.write(data.encode('utf8', 'replace'))

        with open(os.path.join(self.gen_dir, 'cub.unique'), 'wb') as unq_file:
            pickle.dump(np.array(unq_words), unq_file)

        with open(os.path.join(self.gen_dir, 'cub.all'), 'wb') as a_file:
            pickle.dump(occ_register, a_file)

        self._load_vocab()



# Flatten 3D image
import torch
from torch import nn


class Flatten3D(nn.Module):
    def forward(self, x):
        x = x.view(x.size()[0], -1)
        return x


# Unflatten to 3D image
class Unflatten3D(nn.Module):
    def forward(self, x):
        x = x.view(x.size()[0], 64, 4, 4)
        return x


# Increase 2 dims to get a 3D tensor
class Unsqueeze3D(nn.Module):
    def forward(self, x):
        x = x.unsqueeze(-1)
        x = x.unsqueeze(-1)
        return x


# CNNEncoder (the input is fixed to be 64x64xchannels)
class CNNEncoder(nn.Module):
    def __init__(self, z_dim=3, c_dim=1, channels=1):
        super().__init__()
        self.z_dim = z_dim
        self.c_dim = c_dim
        self.pipe = nn.Sequential(
            nn.Conv2d(channels, 32, 4, 2, 1),
            nn.ReLU(True),
            nn.Conv2d(32, 32, 4, 2, 1),
            nn.ReLU(True),
            nn.Conv2d(32, 64, 4, 2, 1),
            nn.ReLU(True),
            nn.Conv2d(64, 64, 4, 2, 1),
            nn.ReLU(True),
            Flatten3D(),
            nn.Linear(4 * 4 * 64, 256),
            nn.ReLU(True)
        )
        # map to shared
        self.S = nn.Linear(256, z_dim)
        # map to private
        self.P = nn.Linear(256, c_dim)

    def forward(self, x):
        tmp = self.pipe(x)
        shared = self.S(tmp)
        private = self.P(tmp)
        return shared, private


# CNNDecoder
class CNNDecoder(nn.Module):
    def __init__(self, z_dim=3, c_dim=1, channels=1):
        super().__init__()
        self.z_dim = z_dim
        self.c_dim = c_dim
        self.pipe = nn.Sequential(
            nn.Linear(z_dim + c_dim, 256),
            nn.ReLU(True),
            nn.Linear(256, 1024),
            nn.ReLU(True),
            Unflatten3D(),
            nn.ConvTranspose2d(64, 64, 4, 2, 1),
            nn.ReLU(True),
            nn.ConvTranspose2d(64, 32, 4, 2, 1),
            nn.ReLU(True),
            nn.ConvTranspose2d(32, 32, 4, 2, 1),
            nn.ReLU(True),
            nn.ConvTranspose2d(32, channels, 4, 2, 1),
        )

    def forward(self, s, p):
        # Reconstruct using both the shared and private
        recons = self.pipe(torch.cat((s, p), 1))
        return recons



# Constants
import torch
from torch import nn

dataSize = torch.Size([3, 32, 32])
imgChans = dataSize[0]
fBase = 32  # base size of filter channels


# Classes
class SVHNEncoder(nn.Module):
    """ Generate latent parameters for SVHN image data. """

    def __init__(self, latent_dim):
        super(SVHNEncoder, self).__init__()
        self.enc = nn.Sequential(
            # input size: 3 x 32 x 32
            nn.Conv2d(imgChans, fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase) x 16 x 16
            nn.Conv2d(fBase, fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 8 x 8
            nn.Conv2d(fBase * 2, fBase * 4, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 4 x 4
        )
        self.c1 = nn.Conv2d(fBase * 4, latent_dim, 4, 1, 0, bias=True)
        self.c2 = nn.Conv2d(fBase * 4, latent_dim, 4, 1, 0, bias=True)
        # c1, c2 size: latent_dim x 1 x 1

    def forward(self, x):
        e = self.enc(x)
        lv = self.c2(e).squeeze()
        return self.c1(e).squeeze(), F.softmax(lv, dim=-1) * lv.size(-1) + Constants.eta


class SVHNDecoder(nn.Module):
    """ Generate a SVHN image given a sample from the latent space. """

    def __init__(self, latent_dim):
        super(SVHNDecoder, self).__init__()
        self.dec = nn.Sequential(
            nn.ConvTranspose2d(latent_dim, fBase * 4, 4, 1, 0, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 4 x 4
            nn.ConvTranspose2d(fBase * 4, fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 8 x 8
            nn.ConvTranspose2d(fBase * 2, fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase) x 16 x 16
            nn.ConvTranspose2d(fBase, imgChans, 4, 2, 1, bias=True),
            nn.Sigmoid()
            # Output size: 3 x 32 x 32
        )

    def forward(self, z):
        z = z.unsqueeze(-1).unsqueeze(-1)  # fit deconv layers
        out = self.dec(z.view(-1, *z.size()[-3:]))
        out = out.view(*z.size()[:-3], *out.size()[1:])
        # consider also predicting the length scale
        return out, torch.tensor(0.75).to(z.device)  # mean, length scale


import os


def test_noisymnist():
    from multiviewdata.torchdatasets import NoisyMNISTDataset
    a = NoisyMNISTDataset(os.getcwd(), download=True)[0]
    assert "index" in a
    assert "views" in a

def test_tangledmnist():
    from multiviewdata.torchdatasets import TangledMNISTDataset
    a = TangledMNISTDataset(os.getcwd(), download=True)[0]
    assert "index" in a
    assert "views" in a

def test_splitmnist():
    from multiviewdata.torchdatasets import SplitMNISTDataset
    a = SplitMNISTDataset(os.getcwd(), download=True)[0]
    assert "index" in a
    assert "views" in a

def test_mfeat():
    from multiviewdata.torchdatasets import MFeatDataset
    a = MFeatDataset(os.getcwd(), download=True)[0]
    assert "index" in a
    assert "views" in a



from torch import nn


class Net(nn.Module):
  def __init__(self, layer_sizes, activ):
    """
    In the constructor we instantiate nn modules and assign them as
    member variables.
    """
    super(Net, self).__init__()
    if activ == 'sigmoid':
        activation = nn.Sigmoid()
    elif activ == 'tanh':
        activation = nn.Tanh()
    else:
        activation = nn.ReLU()
    self.layer1 = nn.Sequential(nn.Linear(layer_sizes[0], layer_sizes[1]), activation)
    self.layer2 = nn.Sequential(nn.Linear(layer_sizes[1], layer_sizes[2]), activation)
    self.layer3 = nn.Sequential(nn.Linear(layer_sizes[2], layer_sizes[3]), activation)
    self.layer4 = nn.Linear(layer_sizes[3], layer_sizes[4], bias=False) # No biases in the final linear layer


  def forward(self, x):
    """
    In the forward function we accept a Variable of input data and we must return
    a Variable of output data. We can use Modules defined in the constructor as
    well as arbitrary operators on Variables.
    """
    output = self.layer1(x)
    output = output.sub(output.mean(0).expand(output.size())) # Removing mean from tensor
    output = self.layer2(output)
    output = output.sub(output.mean(0).expand(output.size())) # Removing mean from tensor
    output = self.layer3(output)
    output = output.sub(output.mean(0).expand(output.size())) # Removing mean from tensor
    output = self.layer4(output)
    output = output.sub(output.mean(0).expand(output.size())) # Removing mean from tensor
    return output


import os

import numpy as np
from torch.utils.data.dataset import Dataset
from torchvision.datasets.utils import download_and_extract_archive


class MFeat(Dataset):
    def __init__(
        self,
        root: str,
        feats: list = None,
        partials: list = None,
        download: bool = False,
    ):
        """

        :param root: Root directory of dataset
        :param feats: Which features to use from ["fac", "fou", "kar", "mor", "pix", "zer"]
        :param partials: Which features to use as partials from ["fac", "fou", "kar", "mor", "pix", "zer"]
        :param download: If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        """
        self.resources = [
            "https://archive.ics.uci.edu/ml/machine-learning-databases/mfeat/mfeat.tar"
        ]
        self.root = root
        if download:
            self.download()
        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )
        if feats is None:
            self.feats = ["fac", "fou", "kar", "mor", "pix", "zer"]
        if partials is None:
            self.partials = None
        self.dataset = dict(
            fac=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-fac")),
            fou=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-fou")),
            kar=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-kar")),
            mor=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-mor")),
            pix=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-pix")),
            zer=np.genfromtxt(os.path.join(self.raw_folder, "mfeat/mfeat-zer")),
        )

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def __getitem__(self, index):
        batch = {"index": index}
        batch["views"] = [self.dataset[feat][index].astype(np.float32) for feat in self.feats]
        if self.partials is not None:
            batch["partials"] = [self.dataset[partial][index].astype(np.float32) for partial in self.partials]
        return batch

    def __len__(self):
        return len(self.dataset["fac"])

    def _check_raw_exists(self) -> bool:
        return os.path.exists(os.path.join(self.raw_folder, "mfeat.tar"))

    def _check_exists(self) -> bool:
        return os.path.exists(os.path.join(self.raw_folder, "mfeat"))

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if not self._check_raw_exists():
            os.makedirs(self.raw_folder, exist_ok=True)
            import ssl

            ssl._create_default_https_context = ssl._create_unverified_context
            # download files
            for url in self.resources:
                filename = url.rpartition("/")[2]
                download_and_extract_archive(
                    url, download_root=self.raw_folder, filename=filename
                )



import math
import os
import zipfile

import h5py
import numpy as np
from torch.utils.data import Dataset


class WIW(Dataset):
    def __init__(self, root, feats=None, partials=None, split="train", download=True):
        """

        :param root: Root directory of dataset
        :param feats: Which features to use from ["eng", "ru", "ger", "it", "vis"]
        :param partials: Which features to use as partials from ["eng", "ru", "ger", "it", "vis"]
        :param split:
        :param download: If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        """
        self.resources = [
            (
                "https://mega.nz/file/Gc0kHBTA#CYpHo_Vs2j1BIML2rlBhxFtOzzAzpkhSIeYT3rE93Go",
                "wiw_data.zip",
            )
        ]
        self.root = root
        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )
        # Dataset parameters

        self.lng_dict = {
            "eng": "english",
            "ger": "german",
            "it": "italian",
            "ru": "russian",
        }
        if feats is None:
            self.feats = ["eng", "ru"]
        if partials is None:
            self.partials = ["vis"]
        self.folder = self.feats[0] + "_" + self.feats[1]
        self.dname = "/wiw_" + self.folder + "_img_pca_300_wv_splitted"  # Dataset Name
        self.dataset = h5py.File(
            os.path.join(self.raw_folder, "wiw_data/")
            + self.folder
            + "/"
            + self.dname
            + ".h5",
            "r",
        )[split]
        self.comment = ""

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def _check_exists(self) -> bool:
        return os.path.exists(os.path.join(self.raw_folder, "wiw_data"))

    def _check_raw_exists(self) -> bool:
        return os.path.exists(os.path.join(self.raw_folder, "wiw_data.zip"))

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        batch = {"index": index.astype(np.float32)}
        batch["views"] = [
            self.dataset["%06d" % index][feat + "_feats"][()].astype(np.float32)
            for feat in self.feats
        ]
        if self.partials is not None:
            batch["partials"] = [
                self.dataset["%06d" % index][partial + "_feats"][()].astype(np.float32)
                for partial in self.partials
            ]
        return batch

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if not self._check_raw_exists():
            os.makedirs(self.raw_folder, exist_ok=True)
            print(
                f"Download raw zip file from https://mega.nz/file/eV0STDTR#w6Xg248RQdzL28VOmoqsFLidJqrlSZKx7f8AGqfA204"
                f"and put it in {self.raw_folder}. This is manual because the python api is extremely slow."
            )
        if self._check_exists():
            return
        with zipfile.ZipFile(os.path.join(self.raw_folder, "wiw_data.zip")) as z:
            z.extractall(self.raw_folder)
        print("Processing...")
        self.process()
        print("Done!")

    def process(self):
        for languages in [("eng", "ger"), ("eng", "it"), ("eng", "ru")]:
            # Split data
            lng1, lng2 = languages
            dataset_name = (
                os.path.join(self.raw_folder, "wiw_data/")
                + lng1
                + "_"
                + lng2
                + "/wiw_"
                + lng1
                + "_"
                + lng2
                + "_img_pca_300_wv.h5"
            )
            dataname = h5py.File(dataset_name, "r")
            dataset_split = dataname["train"]

            samples_size = count_split_size(dataset_split)
            train_size = int(math.ceil(0.7 * samples_size))
            val_size = int(math.ceil(0.15 * samples_size))
            test_size = samples_size - (train_size + val_size)
            splits_size = [train_size, val_size, test_size]

            train_data = []
            val_data = []
            test_data = []

            for idx in range(samples_size):
                tup = dataset_split["%06d" % idx]
                lng1_feats = tup[lng1 + "_feats"][()]
                lng2_feats = tup[lng2 + "_feats"][()]
                vis_feats = tup["vis_feats"][()]
                lng1_desc = tup[lng1 + "_descriptions_feats"][()][0]  # .value[0]
                lng2_desc = tup[lng2 + "_descriptions_feats"][()][0]  # .value[0]

                if idx < train_size:
                    train_data.append(
                        (lng1_feats, lng2_feats, vis_feats, lng1_desc, lng2_desc)
                    )
                elif idx < train_size + val_size:
                    val_data.append(
                        (lng1_feats, lng2_feats, vis_feats, lng1_desc, lng2_desc)
                    )
                else:
                    test_data.append(
                        (lng1_feats, lng2_feats, vis_feats, lng1_desc, lng2_desc)
                    )

            # Write data
            fname = dataset_name[:-3] + "_splitted"
            h5output = h5py.File(fname + ".h5", "w")
            # The HDF5 file will contain a top-level group for each split
            train = h5output.create_group("train")
            val = h5output.create_group("val")
            test = h5output.create_group("test")
            splits = ["train", "val", "test"]
            split_dims = []
            img_dims = train_data[0][2].shape[0]
            txt_dims = [train_data[0][0].shape[0], train_data[0][1].shape[0]]
            data_idx = 0
            for split in splits:
                dims = splits_size[data_idx]
                textual_features = []
                data_dim_idx = 0
                for dim_idx in range(dims):
                    if split == "train":
                        container = train.create_group("%06d" % data_dim_idx)
                    elif split == "val":
                        container = val.create_group("%06d" % data_dim_idx)
                    else:
                        container = test.create_group("%06d" % data_dim_idx)
                    lng_idx = 0
                    if split == "train":
                        for j, lng in enumerate(languages):
                            text_data = container.create_dataset(
                                lng + "_feats", (txt_dims[j],), dtype="float32"
                            )
                            text_data[:] = train_data[dim_idx][j]
                            text_data = container.create_dataset(
                                lng + "_descriptions_feats",
                                (1,),
                                dtype=h5py.special_dtype(vlen=str),
                            )
                            text_data[:] = train_data[dim_idx][3 + j]
                            lng_idx += 1
                        image_data = container.create_dataset(
                            "vis_feats", (img_dims,), dtype="float32"
                        )
                        image_data[:] = train_data[dim_idx][2]
                    elif split == "val":
                        for j, lng in enumerate(languages):
                            text_data = container.create_dataset(
                                lng + "_feats", (txt_dims[j],), dtype="float32"
                            )
                            text_data[:] = val_data[dim_idx][j]
                            text_data = container.create_dataset(
                                lng + "_descriptions_feats",
                                (1,),
                                dtype=h5py.special_dtype(vlen=str),
                            )
                            text_data[:] = val_data[dim_idx][3 + j]
                            lng_idx += 1
                        image_data = container.create_dataset(
                            "vis_feats", (img_dims,), dtype="float32"
                        )
                        image_data[:] = val_data[dim_idx][2]
                    else:
                        for j, lng in enumerate(languages):
                            text_data = container.create_dataset(
                                lng + "_feats", (txt_dims[j],), dtype="float32"
                            )
                            text_data[:] = test_data[dim_idx][j]
                            text_data = container.create_dataset(
                                lng + "_descriptions_feats",
                                (1,),
                                dtype=h5py.special_dtype(vlen=str),
                            )
                            text_data[:] = test_data[dim_idx][3 + j]
                            lng_idx += 1
                        image_data = container.create_dataset(
                            "vis_feats", (img_dims,), dtype="float32"
                        )
                        image_data[:] = test_data[dim_idx][2]
                    data_dim_idx += 1
                data_idx += 1
            h5output.close()


def count_split_size(dataset_split):
    split_size = 0
    for _ in dataset_split:
        split_size += 1
    return split_size



from .cars3d import Cars
from .generic import GenericDataset
from .mnist import NoisyMNIST, SplitMNIST, TangledMNIST, MNISTSVHN
from .wiw import WIW
from .xrmb import XRMB
from .mfeat import MFeat
from .twitter import Twitter
from .cub import Cub




# Constants
import torch
from torch import nn

imgChans = 3
fBase = 64

class ImageEncoder(nn.Module):
    """ Generate latent parameters for CUB image data. """

    def __init__(self, latentDim):
        super(ImageEncoder, self).__init__()
        modules = [
            # input size: 3 x 128 x 128
            nn.Conv2d(imgChans, fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # input size: 1 x 64 x 64
            nn.Conv2d(fBase, fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 32 x 32
            nn.Conv2d(fBase * 2, fBase * 4, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 16 x 16
            nn.Conv2d(fBase * 4, fBase * 8, 4, 2, 1, bias=True),
            nn.ReLU(True)]
        # size: (fBase * 8) x 4 x 4

        self.enc = nn.Sequential(*modules)
        self.c1 = nn.Conv2d(fBase * 8, latentDim, 4, 1, 0, bias=True)
        self.c2 = nn.Conv2d(fBase * 8, latentDim, 4, 1, 0, bias=True)
        # c1, c2 size: latentDim x 1 x 1

    def forward(self, x):
        e = self.enc(x)
        return self.c1(e).squeeze(), F.softplus(self.c2(e)).squeeze() + Constants.eta


class ImageDecoder(nn.Module):
    """ Generate an image given a sample from the latent space. """

    def __init__(self, latentDim):
        super(ImageDecoder, self).__init__()
        modules = [nn.ConvTranspose2d(latentDim, fBase * 8, 4, 1, 0, bias=True),
                   nn.ReLU(True), ]

        modules.extend([
            nn.ConvTranspose2d(fBase * 8, fBase * 4, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 4) x 16 x 16
            nn.ConvTranspose2d(fBase * 4, fBase * 2, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase * 2) x 32 x 32
            nn.ConvTranspose2d(fBase * 2, fBase, 4, 2, 1, bias=True),
            nn.ReLU(True),
            # size: (fBase) x 64 x 64
            nn.ConvTranspose2d(fBase, imgChans, 4, 2, 1, bias=True),
            nn.Sigmoid()
            # Output size: 3 x 128 x 128
        ])
        self.dec = nn.Sequential(*modules)

    def forward(self, z):
        z = z.unsqueeze(-1).unsqueeze(-1)  # fit deconv layers
        out = self.dec(z.view(-1, *z.size()[-3:]))
        out = out.view(*z.size()[:-3], *out.size()[1:])
        return out, torch.tensor(0.01).to(z.device)

# Constants
maxSentLen = 32  # max length of any description for birds dataset
minOccur = 3
embeddingDim = 128
lenWindow = 3
fBase = 32
vocabSize = 1590
vocab_path = '../data/cub/oc:{}_sl:{}_s:{}_w:{}/cub.vocab'.format(minOccur, maxSentLen, 300, lenWindow)


# Classes
class SentenceEncoder(nn.Module):
    """ Generate latent parameters for sentence data. """

    def __init__(self, latentDim):
        super(SentenceEncoder, self).__init__()
        self.embedding = nn.Embedding(vocabSize, embeddingDim, padding_idx=0)
        self.enc = nn.Sequential(
            # input size: 1 x 32 x 128
            nn.Conv2d(1, fBase, 4, 2, 1, bias=False),
            nn.BatchNorm2d(fBase),
            nn.ReLU(True),
            # size: (fBase) x 16 x 64
            nn.Conv2d(fBase, fBase * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(fBase * 2),
            nn.ReLU(True),
            # size: (fBase * 2) x 8 x 32
            nn.Conv2d(fBase * 2, fBase * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # # size: (fBase * 4) x 4 x 16
            nn.Conv2d(fBase * 4, fBase * 4, (1, 4), (1, 2), (0, 1), bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # size: (fBase * 8) x 4 x 8
            nn.Conv2d(fBase * 4, fBase * 4, (1, 4), (1, 2), (0, 1), bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # size: (fBase * 8) x 4 x 4
        )
        self.c1 = nn.Conv2d(fBase * 4, latentDim, 4, 1, 0, bias=False)
        self.c2 = nn.Conv2d(fBase * 4, latentDim, 4, 1, 0, bias=False)
        # c1, c2 size: latentDim x 1 x 1

    def forward(self, x):
        e = self.enc(self.embedding(x.long()).unsqueeze(1))
        mu, logvar = self.c1(e).squeeze(), self.c2(e).squeeze()
        return mu, F.softplus(logvar) + Constants.eta


class SentenceDecoder(nn.Module):
    """ Generate a sentence given a sample from the latent space. """

    def __init__(self, latentDim):
        super(SentenceDecoder, self).__init__()
        self.dec = nn.Sequential(
            nn.ConvTranspose2d(latentDim, fBase * 4, 4, 1, 0, bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # size: (fBase * 8) x 4 x 4
            nn.ConvTranspose2d(fBase * 4, fBase * 4, (1, 4), (1, 2), (0, 1), bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # size: (fBase * 8) x 4 x 8
            nn.ConvTranspose2d(fBase * 4, fBase * 4, (1, 4), (1, 2), (0, 1), bias=False),
            nn.BatchNorm2d(fBase * 4),
            nn.ReLU(True),
            # size: (fBase * 4) x 8 x 32
            nn.ConvTranspose2d(fBase * 4, fBase * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(fBase * 2),
            nn.ReLU(True),
            # size: (fBase * 2) x 16 x 64
            nn.ConvTranspose2d(fBase * 2, fBase, 4, 2, 1, bias=False),
            nn.BatchNorm2d(fBase),
            nn.ReLU(True),
            # size: (fBase) x 32 x 128
            nn.ConvTranspose2d(fBase, 1, 4, 2, 1, bias=False),
            nn.ReLU(True)
            # Output size: 1 x 64 x 256
        )
        # inverts the 'embedding' module upto one-hotness
        self.toVocabSize = nn.Linear(embeddingDim, vocabSize)

    def forward(self, z):
        z = z.unsqueeze(-1).unsqueeze(-1)  # fit deconv layers
        out = self.dec(z.view(-1, *z.size()[-3:])).view(-1, embeddingDim)

        return self.toVocabSize(out).view(*z.size()[:-3], maxSentLen, vocabSize),



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("./requirements.txt", "r") as f:
    REQUIRED_PACKAGES = f.read()

setup(
    name="multiviewdata",
    version="0.2.1",
    packages=find_packages(),
    url="",
    license="",
    author="James Chapman",
    author_email="james.chapman.19@ucl.ac.uk",
    install_requires=REQUIRED_PACKAGES,
    description="Packaged data modules for multiview learning benchmarks",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import os

from scipy.io import loadmat
from torch.utils.data import Dataset
from torchvision.datasets.utils import download_url
import numpy as np

class XRMB(Dataset):
    def __init__(
        self,
        root,
        train=True,
        download=False,
    ):
        """

        :param root: Root directory of dataset
        :param train:
        :param download: If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        """
        self.root = root
        self.resources = [
            ("http://ttic.edu/livescu/XRMB_data/full/XRMBf2KALDI_window7_single1.mat"),
            ("http://ttic.edu/livescu/XRMB_data/full/XRMBf2KALDI_window7_single2.mat"),
        ]
        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found." + " You can use download=True to download it"
            )
        if train:
            view_1, view_2 = (
                loadmat("XRMBf2KALDI_window7_single1.mat")["X1"],
                loadmat("XRMBf2KALDI_window7_single2.mat")["X2"],
            )
        else:
            view_1, view_2 = (
                loadmat("XRMBf2KALDI_window7_single1.mat")["XTe1"],
                loadmat("XRMBf2KALDI_window7_single2.mat")["XTe2"],
            )
        self.dataset = dict(view_1=view_1, view_2=view_2)

    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    def __len__(self):
        return len(self.dataset["view_1"])

    def __getitem__(self, index):
        return {
            "views": (self.dataset["view_1"][index].astype(np.float32), self.dataset["view_2"][index].astype(np.float32)),
            "index": index.astype(np.float32),
        }

    def _check_exists(self) -> bool:
        return os.path.exists(
            os.path.join(self.raw_folder, "XRMBf2KALDI_window7_single1.mat")
        ) and os.path.exists(
            os.path.join(self.raw_folder, "XRMBf2KALDI_window7_single2.mat")
        )

    def download(self) -> None:
        """Download the data if it doesn't exist in processed_folder already."""

        if self._check_exists():
            return

        os.makedirs(self.raw_folder, exist_ok=True)
        import ssl

        ssl._create_default_https_context = ssl._create_unverified_context
        views = []
        for url in self.resources:
            filename = url.rpartition("/")[2]
            views.append(download_url(url, self.raw_folder, filename))
        print("Done!")



import numpy as np
import torch
import torch.utils.data
import torchvision
from torch.utils.data import Dataset
from torchvision import datasets
from torchvision import transforms


class SplitMNIST(Dataset):
    """
    Class to generate paired noisy mnist data
    """

    def __init__(
            self,
            root: str,
            mnist_type: str = "MNIST",
            train: bool = True,
            flatten: bool = True,
            download=False,
    ):
        """
        :param root: Root directory of dataset
        :param mnist_type: "MNIST", "FashionMNIST" or "KMNIST"
        :param train: whether this is train or test
        :param flatten: whether to flatten the data into array or use 2d images
        """

        self.dataset = load_mnist(mnist_type, train, root, download)
        self.flatten = flatten

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        x_a, label = self.dataset[idx]
        x_b = x_a[:, :, 14:]
        x_a = x_a[:, :, :14]
        if self.flatten:
            x_a = torch.flatten(x_a)
            x_b = torch.flatten(x_b)
        return {"views": (x_a, x_b), "label": label, "index": idx}


class NoisyMNIST(Dataset):
    """
    Class to generate paired noisy mnist data
    """

    def __init__(
            self,
            root: str,
            mnist_type: str = "MNIST",
            train: bool = True,
            flatten: bool = True,
            download=False,
    ):
        """
        :param root: Root directory of dataset
        :param mnist_type: "MNIST", "FashionMNIST" or "KMNIST"
        :param train: whether this is train or test
        :param download: If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        :param flatten: whether to flatten the data into array or use 2d images
        """
        self.dataset = load_mnist(mnist_type, train, root, download)
        self.a_transform = torchvision.transforms.RandomRotation((-45, 45))
        self.b_transform = transforms.Compose(
            [
                transforms.Lambda(_add_mnist_noise),
                transforms.Lambda(self.__threshold_func__),
            ]
        )
        self.targets = self.dataset.targets
        self.filtered_classes = []
        self.filtered_nums = []
        for i in range(10):
            self.filtered_nums.append(np.where(self.targets == i)[0])
        self.flatten = flatten

    def __threshold_func__(self, x):
        x[x > 1] = 1
        return x

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        x_a, label = self.dataset[index]
        x_a = self.a_transform(x_a)
        # get random index of image with same class
        random_index = np.random.choice(self.filtered_nums[label])
        x_b = self.b_transform(self.dataset[random_index][0])
        if self.flatten:
            x_a = torch.flatten(x_a)
            x_b = torch.flatten(x_b)
        return {"views": (x_a.float(), x_b.float()), "label": label, "index": index}


class TangledMNIST(Dataset):
    """
    Class to generate paired tangled MNIST dataset
    """

    def __init__(
            self,
            root: str,
            mnist_type: str = "MNIST",
            train: bool = True,
            flatten: bool = True,
            download=False,
    ):
        """
        :param root: Root directory of dataset
        :param mnist_type: "MNIST", "FashionMNIST" or "KMNIST"
        :param train: whether this is train or test
        :param flatten: whether to flatten the data into array or use 2d images
        """
        self.dataset = load_mnist(mnist_type, train, root, download)
        self.transform = torchvision.transforms.RandomRotation((-45, 45))
        self.targets = self.dataset.targets
        self.filtered_classes = []
        self.filtered_nums = []
        for i in range(10):
            self.filtered_nums.append(np.where(self.targets == i)[0])
        self.flatten = flatten

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        x_a, label = self.dataset[idx]
        x_a = self.transform(x_a)
        # get random index of image with same class
        random_index = np.random.choice(self.filtered_nums[label])
        x_b = self.transform(self.dataset[random_index][0])
        if self.flatten:
            x_a = torch.flatten(x_a)
            x_b = torch.flatten(x_b)
        return {"views": (x_a, x_b), "label": label, "index": idx}


def _add_mnist_noise(x):
    x = x + torch.rand(28, 28) / 10
    return x


def load_mnist(mnist_type, train, root, download):
    if mnist_type == "MNIST":
        dataset = datasets.MNIST(
            root,
            train=train,
            download=download,
            transform=torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor()]
            ),
        )

    elif mnist_type == "FashionMNIST":
        dataset = datasets.FashionMNIST(
            root,
            train=train,
            download=download,
            transform=torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor()]
            ),
        )
    elif mnist_type == "KMNIST":
        dataset = datasets.KMNIST(
            root,
            train=train,
            download=download,
            transform=torchvision.transforms.Compose(
                [
                    torchvision.transforms.ToTensor(),
                ]
            ),
        )
    else:
        raise ValueError
    return dataset


class MNISTSVHN(Dataset):
    """
    We construct a dataset of pairs of MNIST and SVHN such that each pair depicts the same digit class. Each instance of a digit class in either dataset is randomly paired with 20 instances of the same digit class from the other dataset.
    """

    def __init__(
            self,
            root: str,
            max_d: int = 10000,
            dm: int = 10,
            train: bool = True,
            download=False,
            flatten: bool = True,
    ):
        """
        :param root: Root directory of dataset
        :param train: whether this is train or test
        :param flatten: whether to flatten the data into array or use 2d images
        """
        self.flatten = flatten
        self.mnist = load_mnist("MNIST", train, root, download)
        self.svhn = self.load_svhn(train, root, download)
        self.mnist_index, self.svhn_index = self.rand_match_on_idx(max_d=max_d, dm=dm)

    def __len__(self):
        return len(self.mnist_index)

    def __getitem__(self, idx):
        x, label = self.mnist[self.mnist_index[idx]]
        y, label_ = self.svhn[self.svhn_index[idx]]
        assert label == label_, "labels not the same so something is wrong with the dataset"
        if self.flatten:
            x = torch.flatten(x)
            y = torch.flatten(y)
        return {"views": (x, y), "label": label, "index": idx}

    def load_svhn(self, train, root, download):
        if train:
            svhn = datasets.SVHN(root, split='train', download=download,transform=torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor()]
            ),)
        else:
            svhn = datasets.SVHN(root, split='test', download=download,transform=torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor()]
            ),)
        svhn.labels = torch.LongTensor(svhn.labels.squeeze().astype(int)) % 10
        return svhn

    def rand_match_on_idx(self, max_d=10000, dm=10):
        """
        l*: sorted labels
        idx*: indices of sorted labels in original list
        """
        mnist_l, mnist_li = self.mnist.targets.sort()
        svhn_l, svhn_li = self.svhn.labels.sort()
        _idx1, _idx2 = [], []
        for l in mnist_l.unique():  # assuming both have same idxs
            l_idx1, l_idx2 = mnist_li[mnist_l == l], svhn_li[svhn_l == l]
            n = min(l_idx1.size(0), l_idx2.size(0), max_d)
            l_idx1, l_idx2 = l_idx1[:n], l_idx2[:n]
            for _ in range(dm):
                _idx1.append(l_idx1[torch.randperm(n)])
                _idx2.append(l_idx2[torch.randperm(n)])
        return torch.cat(_idx1), torch.cat(_idx2)
